@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2018 Artifex Software, Inc.
+/* Copyright (C) 2001-2021 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -24,6 +24,8 @@
 #include "string_.h"
 #include <ctype.h>	/* for isalpha, etc. */
 
+extern dev_proc_open_device(clist_open);
+
 /* Save the current clist state into a saved page structure,
  * and optionally stashes the files into the given save_files
  * pointers.
@@ -106,7 +108,7 @@ params_out:
         }
     }
     /* Now re-open the clist device so that we get new files for the next page */
-    return (*gs_clist_device_procs.open_device) ((gx_device *) pdev);
+    return clist_open((gx_device *) pdev);
 }
 
 /* Save a page. The elements are allocated by this function in non_gc_memory */
@@ -779,7 +781,7 @@ gx_saved_pages_list_print(gx_device_printer *pdev, gx_saved_pages_list *list,
                             break;
                         curr_elem = curr_elem->prev;
                         /* Below is not needed since we never print reverse even/odd */
-                        if (page_skip < -1)
+                        if (page_skip < -1) /* lgtm [cpp/constant-comparison] */
                             curr_elem = curr_elem->prev;
                     }
                     if (curr_elem == NULL) {
