@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2012 Artifex Software, Inc.
+/* Copyright (C) 2001-2022 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  7 Mt. Lassen Drive - Suite A-134, San Rafael,
-   CA  94903, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
+   CA 94945, U.S.A., +1(415)492-9861, for further information.
 */
 
 
@@ -19,6 +19,8 @@
 #ifndef gxdevsop_INCLUDED
 #  define gxdevsop_INCLUDED
 
+#include "gxdevcli.h"
+
 /* This file enumerates a series of device specific operations, that can be
  * performed using the 'dev_spec_op' procedure in the gs_device structure.
  *
@@ -27,7 +29,7 @@
  * ioctl is to unix file handles.
  *
  * Design features of this scheme ensure that:
- *  * Devices that support a given call call efficiently implement both
+ *  * Devices that support a given call can efficiently implement both
  *    input and output.
  *  * Devices that do not support a given call can efficiently refuse it
  *    (without having to know all the possible calls).
@@ -127,6 +129,7 @@ typedef struct dev_param_req_s {
 /* structure used to pass the parameters for pattern handling */
 typedef struct pattern_accum_param_t {
     void *pinst;
+    gs_memory_t *interpreter_memory;
     void *graphics_state;
     int pinst_id;
 }pattern_accum_param_s;
@@ -224,6 +227,8 @@ enum {
      */
     gxdso_is_pdf14_device,
 
+    gxdso_supports_pattern_transparency,
+
     /* gxdso_device_child:
      *      data = pointer to gxdso_device_child_request struct
      *      size = sizeof(gxdso_device_child_request)
@@ -307,6 +312,144 @@ enum {
      * device.
      */
     gxdso_in_pattern_accumulator,
+    /* Determine if we are in a PDF14 device and the target is a separation
+     * device.   In this case, we may want to not use the alternate tint
+     * tranform even if the blending color space is RGB or Gray. */
+    gxdso_pdf14_sep_device,
+    /* Used only by pdfwrite to pass a Form Appearance Name, so that
+     * we can use the name in a pdfmark.
+     */
+    gxdso_pdf_form_name,
+    gxdso_pdf_last_form_ID,
+    /* Restrict the supplied bbox to that actually used by the underlying device.
+     * Used to restrict alphabits drawing to the area defined by compositors etc.*/
+    gxdso_restrict_bbox,
+    /* JPEG passthrough requests/control. Currently used for the pdfwrite family only.
+     */
+    gxdso_JPEG_passthrough_query,
+    gxdso_JPEG_passthrough_begin,
+    gxdso_JPEG_passthrough_data,
+    gxdso_JPEG_passthrough_end,
+    /* And similarly JPX passthrough requests/control. Currently used for the pdfwrite family only.
+     */
+    gxdso_JPX_passthrough_query,
+    gxdso_JPX_passthrough_begin,
+    gxdso_JPX_passthrough_data,
+    gxdso_JPX_passthrough_end,
+
+    gxdso_supports_iccpostrender,
+    /* Retrieve the last device in a device chain
+       (either forwarding or subclass devices).
+     */
+    gxdso_current_output_device,
+    /* Should we call copy_color rather than resolving images to fill_rectangles? */
+    gxdso_copy_color_is_fast,
+    /* gxdso_is_encoding_direct:
+     *     data = NULL
+     *     size = 0
+     * Returns 1 if the device maps color byte values directly into the device,
+     * 0 otherwise.
+     */
+    gxdso_is_encoding_direct,
+    /* gxdso_event_info:
+     *     data = dev_param_req_t
+     *     size = sizeof(dev_param-req_t
+     * Passes a single name in request->Param, naming the event which occurred.
+     * Used to send a warning to pdfwrite that some event has happened we want to know about.
+     * Currently this is used in pdf_font.ps to signal that a substittue font has been
+     * used. If we are emitting PDF/A then we need to abort it, as the Widths array of
+     * the PDF font may not match the widths of the glyphs in the font.
+     */
+    gxdso_event_info,
+    /* gxdso_overprint_active:
+    *     data = NULL
+    *     size = 0
+    * Returns 1 if the overprint device is active,
+    * 0 otherwise.
+    */
+    gxdso_overprint_active,
+    /* gxdso_in_smask:
+    *     data = NULL
+    *     size = 0
+    * Returns 1 if we are within an smask (either construction or usage),
+    * 0 otherwise.
+    */
+    gxdso_in_smask,
+    /* gxdso_in_smask_construction:
+    *     data = NULL
+    *     size = 0
+    * Returns 1 if we are within an smask construction,
+    * 0 otherwise.
+    */
+    gxdso_in_smask_construction,
+    /* Debug only dsos follow here */
+#ifdef DEBUG
+    /* Private dso used to check that a printer device properly forwards to the default */
+    gxdso_debug_printer_check,
+#endif
+    /* gxdso_supports_alpha:
+     *     data = NULL
+     *     size = 0
+     * Returns 1 if the rendering device supports alpha,
+     * 0 otherwise.
+     */
+    gxdso_supports_alpha,
+    /* gxdso_reopen_after_init:
+     *     data = NULL
+     *     size = 0
+     * Returns 1 if the device should be closed/reopened after gs
+     * finishes initialisation (e.g. to give it a chance to fetch
+     * configuration from registered callout handlers),
+     * 0 otherwise.
+     */
+    gxdso_reopen_after_init,
+
+    /* Normally, when we set a device ICC profile, we validate that the number
+     * of components against the number of components the device supports. But
+     * in certain cases, we don't want to do that (DeviceN devices and nullpage,
+     * for example).
+     */
+    gxdso_skip_icc_component_validation,
+
+    /* gxdso_copy_alpha_disabled:
+     *     data = NULL
+     *     size = 0
+     * Returns 1 if the command list device sets clist_disable_copy_alpha flag,
+     * 0 otherwise.
+     */
+    gxdso_copy_alpha_disabled,
+
+    /* gxdso_set_HWSize:
+     *     data = int[2], [0] is width [1] is height
+     *     size = sizeof(int[2])
+     * Returns 1 if the device controls dev->width, dev->height
+     * 0 otherwise.
+     * NB: caller should set dev->wdith and dev->height if return is <= 0
+     */
+    gxdso_set_HWSize,
+
+    /* gxdso_device_insert_child:
+     *     data = pointer to device to insert as child of the device
+     *            handling this special op.
+     *     size = 0
+     * Returns 0 if completed successfully, negative otherwise.
+     */
+    gxdso_device_insert_child,
+
+    /* Determine if a given device is a clist one. Returns 1 if it is. */
+    gxdso_is_clist_device,
+
+    /* Determine if a given device is a null device. Returns 1 if it is. */
+    gxdso_is_null_device,
+
+    /* Get information about pdf14 device overprint simulation state
+     *     data = unsigned char[2], [0] is overprint_sim_push [1] is num_spot_colors_int
+     *     size = sizeof(unsigned char[2])
+     * Returns 1 if returned values are valid
+     * 0 otherwise.
+     */
+    gxdso_overprintsim_state,
+
     /* Add new gxdso_ keys above this. */
     gxdso_pattern__LAST
 };
