@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2018 Artifex Software, Inc.
+/* Copyright (C) 2001-2021 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -59,15 +59,114 @@ iclass_proc(gs_image_class_4_color);
 
 static irender_proc(image_render_color_DeviceN);
 static irender_proc(image_render_color_icc_tpr);
+#ifndef WITH_CAL
 static irender_proc(image_render_color_thresh);
+#else
+
+#include "cal.h"
+
+static irender_proc(image_render_color_ht_cal);
+
+static int
+image_render_color_ht_cal_skip_line(gx_image_enum *penum,
+                                    gx_device     *dev)
+{
+    return !cal_halftone_next_line_required(penum->cal_ht);
+}
+
+static void
+color_halftone_callback(cal_halftone_data_t *ht, void *arg)
+{
+    gx_device *dev = arg;
+    gx_color_index dev_white = gx_device_white(dev);
+    gx_color_index dev_black = gx_device_black(dev);
+
+    if (dev->is_planar) {
+        (*dev_proc(dev, copy_planes)) (dev, ht->data, ht->x + (ht->offset_x<<3), ht->raster,
+            gx_no_bitmap_id, ht->x, ht->y, ht->w, ht->h,
+            ht->plane_raster);
+    } else {
+        (*dev_proc(dev, copy_mono)) (dev, ht->data, ht->x + (ht->offset_x<<3), ht->raster,
+            gx_no_bitmap_id, ht->x, ht->y, ht->w, ht->h, dev_white,
+            dev_black);
+    }
+}
+
+static cal_halftone*
+color_halftone_init(gx_image_enum *penum)
+{
+    cal_halftone *cal_ht = NULL;
+    gx_dda_fixed dda_ht;
+    cal_context *ctx = penum->memory->gs_lib_ctx->core->cal_ctx;
+    int k;
+    gx_ht_order *d_order;
+    int code;
+    byte *cache = (penum->color_cache != NULL ? penum->color_cache->device_contone : NULL);
+    cal_matrix matrix;
+    int clip_x, clip_y;
+
+    if (!gx_device_must_halftone(penum->dev))
+        return NULL;
+
+    if (penum->pgs == NULL || penum->pgs->dev_ht[HT_OBJTYPE_DEFAULT] == NULL)
+        return NULL;
+    dda_ht = penum->dda.pixel0.x;
+    if (penum->dxx > 0)
+        dda_translate(dda_ht, -fixed_epsilon);
+    matrix.xx = penum->matrix.xx;
+    matrix.xy = penum->matrix.xy;
+    matrix.yx = penum->matrix.yx;
+    matrix.yy = penum->matrix.yy;
+    matrix.tx = penum->matrix.tx + matrix.xx * penum->rect.x + matrix.yx * penum->rect.y;
+    matrix.ty = penum->matrix.ty + matrix.xy * penum->rect.x + matrix.yy * penum->rect.y;
+    clip_x = fixed2int(penum->clip_outer.p.x);
+    clip_y = fixed2int(penum->clip_outer.p.y);
+    cal_ht = cal_halftone_init(ctx,
+                               penum->memory->non_gc_memory,
+                               penum->rect.w,
+                               penum->rect.h,
+                               &matrix,
+                               penum->dev->color_info.num_components,
+                               cache,
+                               clip_x,
+                               clip_y,
+                               fixed2int_ceiling(penum->clip_outer.q.x) - clip_x,
+                               fixed2int_ceiling(penum->clip_outer.q.y) - clip_y,
+                               penum->adjust);
+    if (cal_ht == NULL)
+        goto fail;
+
+    for (k = 0; k < penum->pgs->dev_ht[HT_OBJTYPE_DEFAULT]->num_comp; k++) {
+        d_order = &(penum->pgs->dev_ht[HT_OBJTYPE_DEFAULT]->components[k].corder);
+        code = gx_ht_construct_threshold(d_order, penum->dev, penum->pgs, k);
+        if (code < 0)
+            goto fail;
+        if (cal_halftone_add_screen(ctx,
+                                    penum->memory->non_gc_memory,
+                                    cal_ht,
+                                    penum->pgs->dev_ht[HT_OBJTYPE_DEFAULT]->components[k].corder.threshold_inverted,
+                                    penum->pgs->dev_ht[HT_OBJTYPE_DEFAULT]->components[k].corder.width,
+                                    penum->pgs->dev_ht[HT_OBJTYPE_DEFAULT]->components[k].corder.full_height,
+                                    -penum->pgs->screen_phase[k].x,
+                                    -penum->pgs->screen_phase[k].y,
+                                    penum->pgs->dev_ht[HT_OBJTYPE_DEFAULT]->components[k].corder.threshold) < 0)
+            goto fail;
+    }
+
+    return cal_ht;
+
+fail:
+    cal_halftone_fin(cal_ht, penum->memory->non_gc_memory);
+    return NULL;
+}
+#endif
 
 static int image_skip_color_icc_tpr(gx_image_enum *penum, gx_device *dev);
 
-irender_proc_t
-gs_image_class_4_color(gx_image_enum * penum)
+int
+gs_image_class_4_color(gx_image_enum * penum, irender_proc_t *render_fn)
 {
-    bool std_cmap_procs;
-    int code;
+    int code = 0;
 #if USE_FAST_HT_CODE
     bool use_fast_thresh = true;
 #else
@@ -115,17 +214,16 @@ gs_image_class_4_color(gx_image_enum * penum)
        then we will need to use those and go through pixel by pixel instead
        of blasting through buffers.  This is true for example with many of
        the color spaces for CUPs */
-    std_cmap_procs = gx_device_uses_std_cmap_procs(penum->dev, penum->pgs);
     if ( (gs_color_space_get_index(penum->pcs) == gs_color_space_index_DeviceN &&
-        penum->pcs->cmm_icc_profile_data == NULL) || penum->use_mask_color ||
-        !std_cmap_procs) {
-        return &image_render_color_DeviceN;
+        penum->pcs->cmm_icc_profile_data == NULL) || penum->use_mask_color) {
+         *render_fn = &image_render_color_DeviceN;
+         return 0;
     }
 
     /* Set up the link now */
     code = dev_proc(penum->dev, get_profile)(penum->dev, &dev_profile);
     if (code < 0)
-        return NULL;    /* This function does not return errors, best we can do is say 'we can't handle this' */
+        return code;
 
     des_num_comp = gsicc_get_device_profile_comps(dev_profile);
     bpc = penum->dev->color_info.depth / des_num_comp;	/* bits per component */
@@ -172,6 +270,10 @@ gs_image_class_4_color(gx_image_enum * penum)
             penum->use_cie_range = (get_cie_range(penum->pcs) != NULL);
         }
     }
+    if (!gx_device_uses_std_cmap_procs(penum->dev, penum->pgs)) {
+        *render_fn = &image_render_color_DeviceN;
+        return code;
+    }
     if (gx_device_must_halftone(penum->dev) && use_fast_thresh &&
         (penum->posture == image_portrait || penum->posture == image_landscape)
         && penum->image_parent_type == gs_image_type1) {
@@ -189,12 +291,23 @@ gs_image_class_4_color(gx_image_enum * penum)
         if ((bpc == 1) && transfer_is_monotonic &&
             (penum->dev->color_info.num_components == 1 || penum->dev->is_planar) &&
             penum->bps == 8) {
+#ifdef WITH_CAL
+            penum->cal_ht = color_halftone_init(penum);
+            if (penum->cal_ht != NULL)
+            {
+                penum->skip_next_line = image_render_color_ht_cal_skip_line;
+                *render_fn = &image_render_color_ht_cal;
+                return code;
+            }
+#else
             code = gxht_thresh_image_init(penum);
             if (code == 0) {
                  /* NB: transfer function is pickled into the threshold arrray */
                  penum->icc_setup.has_transfer = false;
-                 return &image_render_color_thresh;
+                 *render_fn = &image_render_color_thresh;
+                 return code;
             }
+#endif
         }
     }
     {
@@ -220,10 +333,11 @@ gs_image_class_4_color(gx_image_enum * penum)
         if (code >= 0) {
             penum->tpr_state = data.state;
             penum->skip_next_line = image_skip_color_icc_tpr;
-            return &image_render_color_icc_tpr;
+            *render_fn = &image_render_color_icc_tpr;
+            return code;
         }
     }
-    return NULL;
+    return code;
 }
 
 /* ------ Rendering procedures ------ */
@@ -331,14 +445,14 @@ decode_row(const gx_image_enum *penum, const byte *psrc, int spp, byte *pdes,
 static int
 image_color_icc_prep(gx_image_enum *penum_orig, const byte *psrc, uint w,
                      gx_device *dev, int *spp_cm_out, byte **psrc_cm,
-                     byte **psrc_cm_start, byte **bufend, bool planar_out)
+                     byte **psrc_cm_start, byte **bufend, int *pspan, bool planar_out)
 {
     const gx_image_enum *const penum = penum_orig; /* const within proc */
     const gs_gstate *pgs = penum->pgs;
     bool need_decode = penum->icc_setup.need_decode;
     gsicc_bufferdesc_t input_buff_desc;
     gsicc_bufferdesc_t output_buff_desc;
-    int num_pixels, spp_cm;
+    int spp_cm;
     int spp = penum->spp;
     bool force_planar = false;
     int num_des_comps;
@@ -348,7 +462,6 @@ image_color_icc_prep(gx_image_enum *penum_orig, const byte *psrc, uint w,
     const byte *planar_src;
     byte *planar_des;
     int j, k;
-    int width;
 
     code = dev_proc(dev, get_profile)(dev, &dev_profile);
     if (code < 0) return code;
@@ -356,6 +469,8 @@ image_color_icc_prep(gx_image_enum *penum_orig, const byte *psrc, uint w,
     if (penum->icc_link == NULL) {
         return gs_rethrow(-1, "ICC Link not created during image render color");
     }
+    if (pspan)
+        *pspan = w;
     /* If the link is the identity, then we don't need to do any color
        conversions except for potentially a decode.  Planar out is a special
        case. For now we let the CMM do the reorg into planar.  We will want
@@ -371,11 +486,18 @@ image_color_icc_prep(gx_image_enum *penum_orig, const byte *psrc, uint w,
         *bufend = *psrc_cm + w;
         *psrc_cm_start = NULL;
     } else {
+        int width = w/spp;
+        int span = (width+31)&~31;
         spp_cm = num_des_comps;
-        *psrc_cm = gs_alloc_bytes(pgs->memory,  w * spp_cm/spp,
-                                  "image_color_icc_prep");
-        *psrc_cm_start = *psrc_cm;
-        *bufend = *psrc_cm +  w * spp_cm/spp;
+
+        if (pspan)
+            *pspan = span;
+        /* Put the buffer on a 32 byte memory alignment for SSE/AVX for every
+         * line. Also extra space for 32 byte overrun. */
+        *psrc_cm_start = gs_alloc_bytes(pgs->memory,  span * spp_cm + 64,
+                                        "image_color_icc_prep");
+        *psrc_cm = *psrc_cm_start + ((32 - (intptr_t)(*psrc_cm_start)) & 31);
+        *bufend = *psrc_cm +  span * spp_cm;
         if (penum->icc_link->is_identity) {
             if (!force_planar) {
                 /* decode only. no CM.  This is slow but does not happen that often */
@@ -401,11 +523,10 @@ image_color_icc_prep(gx_image_enum *penum_orig, const byte *psrc, uint w,
                     planar_src = psrc;
                 }
                 /* Now to planar */
-                width = w/spp;
                 planar_des = *psrc_cm;
                 for (k = 0; k < width; k++) {
                     for (j = 0; j < spp; j++) {
-                        *(planar_des + j * width) = *planar_src++;
+                        *(planar_des + j * span) = *planar_src++;
                     }
                     planar_des++;
                 }
@@ -417,18 +538,17 @@ image_color_icc_prep(gx_image_enum *penum_orig, const byte *psrc, uint w,
             }
         } else {
             /* Set up the buffer descriptors. planar out always ends up here */
-            num_pixels = w/spp;
             gsicc_init_buffer(&input_buff_desc, spp, 1,
                           false, false, false, 0, w,
-                          1, num_pixels);
+                          1, width);
             if (!force_planar) {
                 gsicc_init_buffer(&output_buff_desc, spp_cm, 1,
-                              false, false, false, 0, num_pixels * spp_cm,
-                              1, num_pixels);
+                              false, false, false, 0, width * spp_cm,
+                              1, width);
             } else {
                 gsicc_init_buffer(&output_buff_desc, spp_cm, 1,
-                              false, false, true, w/spp, w/spp,
-                              1, num_pixels);
+                              false, false, true, span, span,
+                              1, width);
             }
             /* For now, just blast it all through the link. If we had a significant reduction
                we will want to repack the data first and then do this.  That will be
@@ -446,19 +566,23 @@ image_color_icc_prep(gx_image_enum *penum_orig, const byte *psrc, uint w,
                     decode_row_cie(penum, psrc, spp, psrc_decode,
                                     psrc_decode+w, get_cie_range(penum->pcs));
                 }
-                (penum->icc_link->procs.map_buffer)(dev, penum->icc_link,
+                code = (penum->icc_link->procs.map_buffer)(dev, penum->icc_link,
                                                     &input_buff_desc,
                                                     &output_buff_desc,
                                                     (void*) psrc_decode,
                                                     (void*) *psrc_cm);
                 gs_free_object(pgs->memory, psrc_decode, "image_color_icc_prep");
+                if (code < 0)
+                    return code;
             } else {
                 /* CM only. No decode */
-                (penum->icc_link->procs.map_buffer)(dev, penum->icc_link,
+                code = (penum->icc_link->procs.map_buffer)(dev, penum->icc_link,
                                                     &input_buff_desc,
                                                     &output_buff_desc,
                                                     (void*) psrc,
                                                     (void*) *psrc_cm);
+                if (code < 0)
+                    return code;
             }
         }
     }
@@ -466,9 +590,46 @@ image_color_icc_prep(gx_image_enum *penum_orig, const byte *psrc, uint w,
     return 0;
 }
 
+#ifdef WITH_CAL
+static int
+image_render_color_ht_cal(gx_image_enum *penum, const byte *buffer, int data_x,
+                          uint w, int h, gx_device * dev)
+{
+    const byte *psrc = buffer + data_x;
+    int code = 0;
+    int spp_cm = 0;
+    byte *psrc_cm = NULL, *psrc_cm_start = NULL;
+    byte *bufend = NULL;
+    byte *input[GX_DEVICE_COLOR_MAX_COMPONENTS];    /* to ensure 128 bit boundary */
+    int i;
+    int planestride;
+
+    if (h == 0)
+        return 0;
+
+    /* Get the buffer into the device color space */
+    code = image_color_icc_prep(penum, psrc, w, dev, &spp_cm, &psrc_cm,
+                                &psrc_cm_start,  &bufend, &planestride, true);
+    if (code < 0)
+        return code;
+
+    for (i = 0; i < spp_cm; i++)
+        input[i] = psrc_cm + w*i;
+
+    code = cal_halftone_process_planar(penum->cal_ht, penum->memory->non_gc_memory,
+                                       (const byte * const *)input, color_halftone_callback, dev);
+
+    /* Free cm buffer, if it was used */
+    if (psrc_cm_start != NULL) {
+        gs_free_object(penum->pgs->memory, (byte *)psrc_cm_start,
+                       "image_render_color_thresh");
+    }
+    return code;
+}
+#else
 static int
 image_render_color_thresh(gx_image_enum *penum_orig, const byte *buffer, int data_x,
-                   uint w, int h, gx_device * dev)
+                          uint w, int h, gx_device * dev)
 {
     gx_image_enum *penum = penum_orig; /* const within proc */
     image_posture posture = penum->posture;
@@ -501,7 +662,7 @@ image_render_color_thresh(gx_image_enum *penum_orig, const byte *buffer, int dat
     if (h != 0 && penum->line_size != 0) {      /* line_size == 0, nothing to do */
         /* Get the buffer into the device color space */
         code = image_color_icc_prep(penum, psrc, w, dev, &spp_cm, &psrc_cm,
-                                    &psrc_cm_start,  &bufend, true);
+                                    &psrc_cm_start,  &bufend, &psrc_planestride, true);
         if (code < 0)
             return code;
     } else {
@@ -537,10 +698,10 @@ image_render_color_thresh(gx_image_enum *penum_orig, const byte *buffer, int dat
                 xrun += penum->x_extent.x;
             vdi = penum->hci;
             contone_stride = penum->line_size;
-            offset_threshold = (- (((long)(penum->thresh_buffer)) +
-                                      penum->ht_offset_bits)) & 15;
+            offset_threshold = (- (((int)(intptr_t)(penum->thresh_buffer)) +
+                                   penum->ht_offset_bits)) & 15;
             for (k = 0; k < spp_out; k ++) {
-                offset_contone[k]   = (- (((long)(penum->line)) +
+                offset_contone[k]   = (- (((int)(intptr_t)(penum->line)) +
                                           contone_stride * k +
                                           penum->ht_offset_bits)) & 15;
             }
@@ -564,9 +725,9 @@ image_render_color_thresh(gx_image_enum *penum_orig, const byte *buffer, int dat
             xrun = dda_current(dda_ht);            /* really yrun, but just used here for landscape */
             dest_height = gxht_dda_length(&dda_ht, src_size);
             data_length = dest_height;
-            offset_threshold = (-(long)(penum->thresh_buffer)) & 15;
+            offset_threshold = (-(int)(intptr_t)(penum->thresh_buffer)) & 15;
             for (k = 0; k < spp_out; k ++) {
-                offset_contone[k]   = (- ((long)(penum->line) +
+                offset_contone[k]   = (- ((int)(intptr_t)(penum->line) +
                                           contone_stride * k)) & 15;
             }
             /* In the landscaped case, we want to accumulate multiple columns
@@ -641,16 +802,16 @@ image_render_color_thresh(gx_image_enum *penum_orig, const byte *buffer, int dat
                                     *(devc_contone_gray+1) = *psrc_temp;
                             }
                         } else {
-                        /* Mono case, forward */
-                        psrc_temp = psrc_cm;
-                        for (k=0; k<src_size; k++) {
-                            dda_next(dda_ht);
-                            xn = fixed2int_var_rounded(dda_current(dda_ht));
-                            while (xr < xn) {
-                                *devc_contone_gray++ = *psrc_temp;
-                                xr++;
-                            }           /* at loop exit xn will be >= xr */
-                            psrc_temp++;
+                            /* Mono case, forward */
+                            psrc_temp = psrc_cm;
+                            for (k=0; k<src_size; k++) {
+                                dda_next(dda_ht);
+                                xn = fixed2int_var_rounded(dda_current(dda_ht));
+                                while (xr < xn) {
+                                    *devc_contone_gray++ = *psrc_temp;
+                                    xr++;
+                                }           /* at loop exit xn will be >= xr */
+                                psrc_temp++;
                             }
                         }
                     } else {
@@ -665,7 +826,7 @@ image_render_color_thresh(gx_image_enum *penum_orig, const byte *buffer, int dat
                                 xr--;
                             }           /* at loop exit xn will be >= xr */
                             psrc_temp++;
-                            }
+                        }
                     }
                     break;
                 /* Monochrome landscape */
@@ -808,10 +969,9 @@ image_render_color_thresh(gx_image_enum *penum_orig, const byte *buffer, int dat
                             dda_next(dda_ht);
                             xn = fixed2int_var_rounded(dda_current(dda_ht));
                             while (xr > xn) {
-                                for (j = 0; j < spp_out; j++) {
+                                for (j = 0; j < spp_out; j++)
                                     *(devc_contone[j] + position) = (psrc_plane[j])[i];
-                                    position -= LAND_BITS;
-                                }
+                                position -= LAND_BITS;
                                 xr--;
                             }           /* at loop exit xn will be <= xr */
                             i++;
@@ -822,9 +982,8 @@ image_render_color_thresh(gx_image_enum *penum_orig, const byte *buffer, int dat
                            and 2x scaling which we will run into in 300 and
                            600dpi devices and content */
                         /* Apply initial offset */
-                        for (k = 0; k < spp_out; k++) {
+                        for (k = 0; k < spp_out; k++)
                             devc_contone[k] = devc_contone[k] + position;
-                        }
                         if (src_size == dest_height) {
                             for (k = 0; k < data_length; k++) {
                                 /* Is it better to unwind this?  We know it is 4 */
@@ -850,10 +1009,9 @@ image_render_color_thresh(gx_image_enum *penum_orig, const byte *buffer, int dat
                                 dda_next(dda_ht);
                                 xn = fixed2int_var_rounded(dda_current(dda_ht));
                                 while (xr > xn) {
-                                    for (j = 0; j < spp_out; j++) {
+                                    for (j = 0; j < spp_out; j++)
                                         *(devc_contone[j] + position) = (psrc_plane[j])[i];
-                                        position -= LAND_BITS;
-                                    }
+                                    position -= LAND_BITS;
                                     xr--;
                                 }           /* at loop exit xn will be <= xr */
                                 i++;
@@ -881,7 +1039,7 @@ flush:
     thresh_align = penum->thresh_buffer + offset_threshold;
     code = gxht_thresh_planes(penum, xrun, dest_width, dest_height,
                               thresh_align, dev, offset_contone,
-                               contone_stride);
+                              contone_stride);
     /* Free cm buffer, if it was used */
     if (psrc_cm_start != NULL) {
         gs_free_object(penum->pgs->memory, (byte *)psrc_cm_start,
@@ -889,9 +1047,10 @@ flush:
     }
     return code;
 }
+#endif
 
 /* Render a color image with 8 or fewer bits per sample using ICC profile. */
-static int 
+static int
 image_skip_color_icc_tpr(gx_image_enum *penum, gx_device *dev)
 {
     transform_pixel_region_data data;
@@ -920,7 +1079,7 @@ image_render_color_icc_tpr(gx_image_enum *penum_orig, const byte *buffer, int da
     if (h == 0)
         return 0;
     code = image_color_icc_prep(penum_orig, psrc, w, dev, &spp_cm, &psrc_cm,
-                                &psrc_cm_start, &bufend, false);
+                                &psrc_cm_start, &bufend, NULL, false);
     if (code < 0) return code;
     psrc_cm_initial = psrc_cm;
     gx_get_cmapper(&cmapper, pgs, dev, has_transfer, must_halftone, gs_color_select_source);
@@ -979,6 +1138,14 @@ image_render_color_DeviceN(gx_image_enum *penum_orig, const byte *buffer, int da
     bits32 test = penum->mask_color.test;
     bool lab_case = false;
 
+    if (device_encodes_tags(dev)) {
+        devc1.tag = (dev->graphics_type_tag & ~GS_DEVICE_ENCODES_TAGS);
+        devc2.tag = (dev->graphics_type_tag & ~GS_DEVICE_ENCODES_TAGS);
+    } else {
+        devc1.tag = 0;
+        devc2.tag = 0;
+    }
+
     if (h == 0)
         return 0;
 
@@ -1057,20 +1224,24 @@ image_render_color_DeviceN(gx_image_enum *penum_orig, const byte *buffer, int da
             dmputs(dev->memory, "\n");
         }
 #endif
-        mcode = remap_color(&cc, pcs, pdevc_next, pgs, dev,
-                           gs_color_select_source);
+        if (lab_case || penum->icc_link == NULL || pcs->cmm_icc_profile_data == NULL)
+            mcode = remap_color(&cc, pcs, pdevc_next, pgs, dev, gs_color_select_source);
+        else
+            mcode = gx_remap_ICC_with_link(&cc, pcs, pdevc_next, pgs, dev,
+                                           gs_color_select_source, penum->icc_link);
+
 mapped:	if (mcode < 0)
             goto fill;
         if (sizeof(pdevc_next->colors.binary.color[0]) <= sizeof(ulong))
             if_debug7m('B', penum->memory,
-                       "[B]0x%x,0x%x,0x%x,0x%x -> 0x%lx,0x%lx,0x%lx\n",
+                       "[B]0x%x,0x%x,0x%x,0x%x -> 0x%lx,0x%lx," PRI_INTPTR "\n",
                        next.v[0], next.v[1], next.v[2], next.v[3],
                        (ulong)pdevc_next->colors.binary.color[0],
                        (ulong)pdevc_next->colors.binary.color[1],
-                       (ulong) pdevc_next->type);
+                       (intptr_t)pdevc_next->type);
         else
             if_debug9m('B', penum->memory,
-                       "[B]0x%x,0x%x,0x%x,0x%x -> 0x%08lx%08lx,0x%08lx%08lx,0x%lx\n",
+                       "[B]0x%x,0x%x,0x%x,0x%x -> 0x%08lx%08lx,0x%08lx%08lx," PRI_INTPTR "\n",
                        next.v[0], next.v[1], next.v[2], next.v[3],
                        (ulong)(pdevc_next->colors.binary.color[0] >>
                                8 * (sizeof(pdevc_next->colors.binary.color[0]) - sizeof(ulong))),
@@ -1078,7 +1249,7 @@ mapped:	if (mcode < 0)
                        (ulong)(pdevc_next->colors.binary.color[1] >>
                                8 * (sizeof(pdevc_next->colors.binary.color[1]) - sizeof(ulong))),
                        (ulong)pdevc_next->colors.binary.color[1],
-                       (ulong) pdevc_next->type);
+                       (intptr_t)pdevc_next->type);
         /* NB: printf above fails to account for sizeof gx_color_index 4 or 8 bytes */
         if (posture != image_skewed && dev_color_eq(*pdevc, *pdevc_next))
             goto set;
