@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2013 Artifex Software, Inc.
+/* Copyright (C) 2001-2021 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -7,8 +7,8 @@
    This software is distributed under license and may not be copied, modified
    or distributed except as expressly authorized under the terms of that
    license.  Refer to licensing information at http://www.artifex.com/
-   or contact Artifex Software, Inc.,  7 Mt. Lassen Drive - Suite A-134,
-   San Rafael, CA  94903, U.S.A., +1(415)492-9861, for further information.
+   or contact Artifex Software, Inc.,  1305 Grant Avenue - Suite 200,
+   Novato, CA 94945, U.S.A., +1(415)492-9861, for further information.
 */
 
 /* psdcmykog device.
@@ -170,6 +170,18 @@ cmykog_open(gx_device * pdev)
   dev->color_info.separable_and_linear = GX_CINFO_SEP_LIN;
   dev->icc_struct->supports_devn = true;
 
+  /* This device is a DeviceN color based device but unlike psdcmyk and
+     tiffsep it can't change its color characteristics.  It remains as
+     a 6 color cmykog device.  Postsript files can change the maxseparations
+     (example Bug 696600 29-07E.PS)  The gdevdevn code will, if it
+     detects a change occured, close the device and reopen (see
+     devn_printer_put_params) possibly resetting the color info.  tiffsep
+     and psdcmyk, when reopened, will  reset their color characteristics based upon
+     the content of the page, avoiding any issues upon the reopen.  Here we
+     will need to make sure to set the color information to avoid problems. */
+  pdev->color_info.max_components = 6;
+  pdev->color_info.depth = pdev->color_info.num_components * 8;
+
   /* For the planar device we need to set up the bit depth of each plane.
    * For other devices this is handled in check_device_separable where
    * we compute the bit shift for the components etc. */
@@ -213,7 +225,10 @@ cmykog_dev_spec_op(gx_device *dev_, int op, void *data, int datasize)
       if (code != gs_error_undefined)
           return code;
   }
-  return gx_default_dev_spec_op(dev_, op, data, datasize);
+  if (op == gxdso_supports_devn || op == gxdso_skip_icc_component_validation) {
+    return true;
+  }
+  return gdev_prn_dev_spec_op(dev_, op, data, datasize);
 }
 
 /*
@@ -279,7 +294,7 @@ cmykog_put_params(gx_device * pdev, gs_param_list * plist)
  * passed to every process_page function:
  */
 typedef struct cmykog_process_arg_s {
-  FILE *spot_file[GX_DEVICE_COLOR_MAX_COMPONENTS];
+  gp_file *spot_file[GX_DEVICE_COLOR_MAX_COMPONENTS];
   char spot_name[GX_DEVICE_COLOR_MAX_COMPONENTS][gp_file_name_sizeof];
   int dev_raster;
 } cmykog_process_arg_t;
@@ -507,7 +522,7 @@ cmykog_process(void *arg_, gx_device *dev_, gx_device *bdev, const gs_int_rect *
   my_rect.p.y = 0;
   my_rect.q.x = w;
   my_rect.q.y = h;
-  code = dev_proc(bdev, get_bits_rectangle)(bdev, &my_rect, &buffer->params, NULL);
+  code = dev_proc(bdev, get_bits_rectangle)(bdev, &my_rect, &buffer->params);
   if (code < 0)
     return code;
 
@@ -536,11 +551,11 @@ cmykog_process(void *arg_, gx_device *dev_, gx_device *bdev, const gs_int_rect *
 }
 
 static void
-write_plane(FILE *file, byte *data, int w, int h, int raster)
+write_plane(gp_file *file, byte *data, int w, int h, int raster)
 {
 #ifndef NO_OUTPUT
   for (; h > 0; h--) {
-      fwrite(data, 1, w, file);
+      gp_fwrite(data, 1, w, file);
       data += raster;
   }
 #endif
@@ -558,7 +573,7 @@ static const char empty[64] = {
 };
 
 static void
-write_empty_plane(FILE *file, int w, int h)
+write_empty_plane(gp_file *file, int w, int h)
 {
 #ifndef NO_OUTPUT
   w *= h;
@@ -566,7 +581,7 @@ write_empty_plane(FILE *file, int w, int h)
     h = w;
     if (h > 64)
         h = 64;
-    fwrite(empty, 1, h, file);
+    gp_fwrite(empty, 1, h, file);
     w -= h;
   }
 #endif
@@ -601,7 +616,7 @@ cmykog_output(void *arg_, gx_device *dev_, void *buffer_)
 /* This, finally is the main entry point that triggers printing for the
  * device. */
 static int
-cmykog_print_page(gx_device_printer * pdev, FILE * prn_stream)
+cmykog_print_page(gx_device_printer * pdev, gp_file * prn_stream)
 {
   gx_device_cmykog * pdevn = (gx_device_cmykog *) pdev;
   int ncomp = pdevn->color_info.num_components;
@@ -671,10 +686,10 @@ cmykog_print_page(gx_device_printer * pdev, FILE * prn_stream)
     for (i = 1; i < ncomp; i++) {
       char tmp[4096];
       int n;
-      fseek(arg->spot_file[i], 0, SEEK_SET);
-      while (!feof(arg->spot_file[i])) {
-        n = fread(tmp, 1, 4096, arg->spot_file[i]);
-        fwrite(tmp, 1, n, prn_stream);
+      gp_fseek(arg->spot_file[i], 0, SEEK_SET);
+      while (!gp_feof(arg->spot_file[i])) {
+        n = gp_fread(tmp, 1, 4096, arg->spot_file[i]);
+        gp_fwrite(tmp, 1, n, prn_stream);
       }
     }
     /* If Ghostscript knows that all 6 planes aren't used, it may
@@ -690,7 +705,7 @@ prn_done:
     /* Close the temporary files. */
     for(i = 1; i < ncomp; i++) {
       if (arg->spot_file[i] != NULL)
-        fclose(arg->spot_file[i]);
+        gp_fclose(arg->spot_file[i]);
       if(arg->spot_name[i][0])
         unlink(arg->spot_name[i]);
     }
@@ -703,73 +718,21 @@ prn_done:
 
 /* Finally, the device definition itself */
 
-#define device_procs(get_color_mapping_procs)\
-{       cmykog_open,			/* open device */\
-        gx_default_get_initial_matrix,	/* initialize matrix */\
-        NULL,                           /* sync_output */\
-        gdev_prn_bg_output_page,        /* output_page */\
-        cmykog_close,                   /* close */\
-        NULL,                           /* map_rgb_color - not used */\
-        NULL,                           /* map_color_rgb - not used */\
-        NULL,                           /* fill_rectangle */\
-        NULL,                           /* tile_rectangle */\
-        NULL,                           /* copy_mono */\
-        NULL,                           /* copy_color */\
-        NULL,                           /* draw_line */\
-        NULL,                           /* get_bits */\
-        gx_devn_prn_get_params,         /* get_params */\
-        cmykog_put_params,              /* put_params */\
-        NULL,                           /* map_cmyk_color - not used */\
-        NULL,                           /* get_xfont_procs */\
-        NULL,                           /* get_xfont_device */\
-        NULL,                           /* map_rgb_alpha_color */\
-        gx_page_device_get_page_device, /* get_page_device */\
-        NULL,                           /* get_alpha_bits */\
-        NULL,                           /* copy_alpha */\
-        NULL,                           /* get_band */\
-        NULL,                           /* copy_rop */\
-        NULL,                           /* fill_path */\
-        NULL,                           /* stroke_path */\
-        NULL,                           /* fill_mask */\
-        NULL,                           /* fill_trapezoid */\
-        NULL,                           /* fill_parallelogram */\
-        NULL,                           /* fill_triangle */\
-        NULL,                           /* draw_thin_line */\
-        NULL,                           /* begin_image */\
-        NULL,                           /* image_data */\
-        NULL,                           /* end_image */\
-        NULL,                           /* strip_tile_rectangle */\
-        NULL,                           /* strip_copy_rop */\
-        NULL,                           /* get_clipping_box */\
-        NULL,                           /* begin_typed_image */\
-        NULL,                           /* get_bits_rectangle */\
-        NULL,                           /* map_color_rgb_alpha */\
-        NULL,                           /* create_compositor */\
-        NULL,                           /* get_hardware_params */\
-        NULL,                           /* text_begin */\
-        NULL,                           /* finish_copydevice */\
-        NULL,                           /* begin_transparency_group */\
-        NULL,                           /* end_transparency_group */\
-        NULL,                           /* begin_transparency_mask */\
-        NULL,                           /* end_transparency_mask */\
-        NULL,                           /* discard_transparency_layer */\
-        gx_devn_prn_get_color_mapping_procs,/* get_color_mapping_procs */\
-        cmykog_get_color_comp_index,    /* get_color_comp_index */\
-        gx_devn_prn_encode_color,       /* encode_color */\
-        gx_devn_prn_decode_color,       /* decode_color */\
-        NULL,                           /* pattern_manage */\
-        NULL,                           /* fill_rectangle_hl_color */\
-        NULL,				/* include_color_space */\
-        NULL,				/* fill_linear_color_scanline */\
-        NULL,				/* fill_linear_color_trapezoid */\
-        NULL,				/* fill_linear_color_triangle */\
-        NULL,                           /* update_spot_equivalent_colors */\
-        gx_devn_prn_ret_devn_params,    /* ret_devn_params */\
-        NULL,                           /* fillpage */\
-        NULL,                           /* push_transparency_state */\
-        NULL,                           /* pop_transparency_state */\
-        NULL,                           /* put_image */\
-        cmykog_dev_spec_op                 /* dev_spec_op */\
+static void
+psdcmykog_initialize_device_procs(gx_device *dev)
+{
+    gdev_prn_initialize_device_procs_bg(dev);
+
+    set_dev_proc(dev, open_device, cmykog_open);
+    set_dev_proc(dev, close_device, cmykog_close);
+    set_dev_proc(dev, get_params, gx_devn_prn_get_params);
+    set_dev_proc(dev, put_params, cmykog_put_params);
+    set_dev_proc(dev, get_color_mapping_procs, gx_devn_prn_get_color_mapping_procs);
+    set_dev_proc(dev, get_color_comp_index, cmykog_get_color_comp_index);
+    set_dev_proc(dev, encode_color, gx_devn_prn_encode_color);
+    set_dev_proc(dev, decode_color, gx_devn_prn_decode_color);
+    set_dev_proc(dev, ret_devn_params, gx_devn_prn_ret_devn_params);
+    set_dev_proc(dev, dev_spec_op, cmykog_dev_spec_op);
 }
 
 fixed_colorant_name DevCMYKOGComponents[] = {
@@ -804,11 +767,9 @@ fixed_colorant_name DevCMYKOGComponents[] = {
 /*
  * PSDCMYKOG 8bits
  */
-static const gx_device_procs cmykog_procs = device_procs(get_cmykog_spot_color_mapping_procs);
-
 const gx_device_cmykog gs_psdcmykog_device =
 {
-  CMYKOG_DEVICE(cmykog_procs, "psdcmykog", 6, GX_CINFO_POLARITY_SUBTRACTIVE, 48, 255, 255, "DeviceCMYK", 600, 600),
+  CMYKOG_DEVICE(psdcmykog_initialize_device_procs, "psdcmykog", 6, GX_CINFO_POLARITY_SUBTRACTIVE, 48, 255, 255, "DeviceCMYK", 600, 600),
   /* device specific parameters */
   { 8,                        /* Bits per color - must match ncomp, depth, etc. above */
     DevCMYKOGComponents,      /* Names of color model colorants */
