@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2018 Artifex Software, Inc.
+/* Copyright (C) 2001-2021 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -24,80 +24,17 @@
    PDF_FOR_OPDFREAD - an integer 0 (false) or 1 (true).
  */
 
+#ifdef PDF_DEVICE_NAME
+
 const gx_device_pdf PDF_DEVICE_IDENT =
-{std_device_dci_type_body(gx_device_pdf, 0, PDF_DEVICE_NAME,
+{std_device_dci_type_body(gx_device_pdf, pdfwrite_initialize_device_procs,
+                          PDF_DEVICE_NAME,
                           &st_device_pdfwrite,
                           DEFAULT_WIDTH_10THS * X_DPI / 10,
                           DEFAULT_HEIGHT_10THS * Y_DPI / 10,
                           X_DPI, Y_DPI,
                           3, 24, 255, 255, 256, 256),
- {pdf_open,
-  gx_upright_get_initial_matrix,
-  NULL,				/* sync_output */
-  pdf_output_page,
-  pdf_close,
-  gx_default_rgb_map_rgb_color,
-  gx_default_rgb_map_color_rgb,
-  gdev_pdf_fill_rectangle,
-  NULL,				/* tile_rectangle */
-  gdev_pdf_copy_mono,
-  gdev_pdf_copy_color,
-  NULL,				/* draw_line */
-  psdf_get_bits,		/* get_bits */
-  gdev_pdf_get_params,
-  gdev_pdf_put_params,
-  NULL,				/* map_cmyk_color */
-  NULL,				/* get_xfont_procs */
-  NULL,				/* get_xfont_device */
-  NULL,				/* map_rgb_alpha_color */
-  gx_page_device_get_page_device,
-  NULL,				/* get_alpha_bits */
-  NULL,				/* copy_alpha */
-  NULL,				/* get_band */
-  NULL,				/* copy_rop */
-  gdev_pdf_fill_path,
-  gdev_pdf_stroke_path,
-  gdev_pdf_fill_mask,
-  NULL,				/* fill_trapezoid */
-  NULL,				/* fill_parallelogram */
-  NULL,				/* fill_triangle */
-  NULL,				/* draw_thin_line */
-  NULL,				/* begin_image */
-  NULL,				/* image_data */
-  NULL,				/* end_image */
-  gdev_pdf_strip_tile_rectangle,
-  NULL,				/* strip_copy_rop */
-  NULL,				/* get_clipping_box */
-  gdev_pdf_begin_typed_image,
-  psdf_get_bits_rectangle,	/* get_bits_rectangle */
-  NULL,				/* map_color_rgb_alpha */
-  gdev_pdf_create_compositor,	/* create_compositor */
-  NULL,				/* get_hardware_params */
-  gdev_pdf_text_begin,
-  NULL,				/* finish_copydevice */
-  gdev_pdf_begin_transparency_group,	/* begin_transparency_group */
-  gdev_pdf_end_transparency_group,	/* end_transparency_group */
-  gdev_pdf_begin_transparency_mask,	/* begin_transparency_mask */
-  gdev_pdf_end_transparency_mask,	/* end_transparency_mask */
-  NULL,	            /* discard_transparency_layer */
-  NULL,				/* get_color_mapping_procs */
-  NULL,				/* get_color_comp_index */
-  NULL,				/* encode_color */
-  NULL,				/* decode_color */
-  NULL, 			/* pattern_manage */
-  gdev_pdf_fill_rectangle_hl_color, 	/* fill_rectangle_hl_color */
-  gdev_pdf_include_color_space, 	/* include_color_space */
-  NULL,				/* fill_linear_color_scanline */
-  NULL,				/* fill_linear_color_trapezoid */
-  NULL,				/* fill_linear_color_triangle */
-  NULL,				/* update_spot_equivalent_colors */
-  NULL,				/* ret_devn_params */
-  gdev_pdf_fillpage,		/* fillpage */
-  NULL,				/* push_transparency_state */
-  NULL,				/* pop_transparency_state */
-  NULL,				/* put_image */
-  gdev_pdf_dev_spec_op		/* dev_spec_op */
- },
+ { 0 },
  psdf_initial_values(PSDF_VERSION_INITIAL, 0 /*false */ ),  /* (!ASCII85EncodePages) */
  0,                     /* pdf_font_dir */
  PDF_FOR_OPDFREAD,		/* is_ps2write */
@@ -105,11 +42,11 @@ const gx_device_pdf PDF_DEVICE_IDENT =
  -1,				/* EndPage */
  1,				/* StartPage */
  1 /*true*/,			/* Optimize */
- 0 /*false*/,			/* ParseDSCCommentsForDocInfo */
+ 1 /*true*/,			/* ParseDSCCommentsForDocInfo */
  1 /*true*/,			/* ParseDSCComments */
  0 /*false*/,			/* EmitDSCWarnings */
  0 /*false*/,			/* CreateJobTicket */
- 0 /*false*/,			/* PreserveEPSInfo */
+ 1 /*true*/,			/* PreserveEPSInfo */
  1 /*true*/,			/* AutoPositionEPSFiles */
  1 /*true*/,			/* PreserveCopyPage */
  0 /*false*/,			/* UsePrologue */
@@ -117,6 +54,8 @@ const gx_device_pdf PDF_DEVICE_IDENT =
  {0,0},				/* PDFXTrimBoxToMediaBoxOffset */
  {0,0},				/* PDFXBleedBoxToTrimBoxOffset */
  1 /* true */,			/* PDFXSetBleedBoxToMediaBox */
+ "",                            /* ocr_language */
+ 0,                             /* ocr_engine */
  1 /*true*/,			/* ReAssignCharacters */
  1 /*true*/,			/* ReEncodeCharacters */
  1,				/* FirstObjectNumber */
@@ -160,9 +99,9 @@ const gx_device_pdf PDF_DEVICE_IDENT =
  {-1, -1},			/* page_dsc_info */
  0 /*false*/,			/* fill_overprint */
  0 /*false*/,			/* stroke_overprint */
+ 1 /* Absolute Colorimetric */, /* rendering intent */
  0 /*false*/,                   /* remap_fill_coilor */
  0 /*false*/,                   /* remap_stroke_coilor */
- 0,				/* overprint_mode */
  gs_no_id,			/* halftone_id */
  {gs_no_id, gs_no_id, gs_no_id, gs_no_id}, /* transfer_ids */
  0,				/* transfer_not_identity */
@@ -233,6 +172,7 @@ const gx_device_pdf PDF_DEVICE_IDENT =
  0,             /* HighLevelForm */
  0,				/* PatternDepth */
  {0,0,0,0,0,0},                 /* AccumulatedPatternMatrix */
+ 0,             /* PatternsSinceForm */
  0,				/* substream_Resources */
  1,				/* pcm_color_info_index == DeviceRGB */
  false,				/* skip_colors */
@@ -244,6 +184,7 @@ const gx_device_pdf PDF_DEVICE_IDENT =
  {{0, 0}, {0, 0}},  /* Charproc BBox */
  0,				/* charproc_just_accumulated */
  1,             /* PS_accumulator */
+ 0,             /* Scaled_accumulator */
  0,				/* accumulating_a_global_object */
  0,				/* cgp */
  0,				/* substituted_pattern_count */
@@ -251,6 +192,7 @@ const gx_device_pdf PDF_DEVICE_IDENT =
  0,				/* image_mask_id */
  0,				/* image_mask_is_SMask */
  0,				/* image_mask_skip */
+ 0,             /* smask_construction */
  0,				/* image_with_SMask */
  {0,0,0,0,0,0}, 		/* gs_matrix converting_image_matrix */
  0,				/* image_mask_scale */
@@ -271,6 +213,7 @@ const gx_device_pdf PDF_DEVICE_IDENT =
  true,				/* DetectDuplicateImages */
  false,				/* AllowIncrementalCFF */
  !PDF_FOR_OPDFREAD,		/* WantsToUnicode */
+ !PDF_FOR_OPDFREAD,		/* PdfmarkCapable */
  !PDF_FOR_OPDFREAD,		/* WantsPageLabels */
  PDF_FOR_OPDFREAD,		/* AllowPSRepeatFunctions */
  true,				/* IsDistiller (true even for ps2write!) */
@@ -285,5 +228,18 @@ const gx_device_pdf PDF_DEVICE_IDENT =
  -1,                    /* Last Form ID, start with -1 which means 'none' */
  0,                     /* ExtensionMetadata */
  0,                     /* PDFFormName */
- 0                      /* PassThroughWriter */
+ 0,                     /* PassThroughWriter */
+ 1.0,                   /* UserUnit */
+ 0,                     /* UseOCR */
+ NULL,                  /* OCRSaved */
+ 0,                     /* OCRStage */
+ NULL,                  /* OCRUnicode */
+ 0,                     /* OCR_char_code */
+ 0,                     /* OCR_glyph */
+ NULL,                  /* ocr_glyphs */
+ 0                   /* initial_pattern_state */
 };
+
+#else
+int dummy;
+#endif
